<?php
/***********************************************************************
| Cerb(tm) developed by Webgroup Media, LLC.
|-----------------------------------------------------------------------
| All source code & content (c) Copyright 2002-2015, Webgroup Media LLC
|   unless specifically noted otherwise.
|
| This source code is released under the Devblocks Public License.
| The latest version of this license can be found here:
| http://cerberusweb.com/license
|
| By using this software, you acknowledge having read this license
| and agree to be bound thereby.
| ______________________________________________________________________
|	http://www.cerbweb.com	    http://www.webgroupmedia.com/
***********************************************************************/
/*
 * IMPORTANT LICENSING NOTE from your friends on the Cerb Development Team
 *
 * Sure, it would be so easy to just cheat and edit this file to use the
 * software without paying for it.  But we trust you anyway.  In fact, we're
 * writing this software for you!
 *
 * Quality software backed by a dedicated team takes money to develop.  We
 * don't want to be out of the office bagging groceries when you call up
 * needing a helping hand.  We'd rather spend our free time coding your
 * feature requests than mowing the neighbors' lawns for rent money.
 *
 * We've never believed in hiding our source code out of paranoia over not
 * getting paid.  We want you to have the full source code and be able to
 * make the tweaks your organization requires to get more done -- despite
 * having less of everything than you might need (time, people, money,
 * energy).  We shouldn't be your bottleneck.
 *
 * We've been building our expertise with this project since January 2002.  We
 * promise spending a couple bucks [Euro, Yuan, Rupees, Galactic Credits] to
 * let us take over your shared e-mail headache is a worthwhile investment.
 * It will give you a sense of control over your inbox that you probably
 * haven't had since spammers found you in a game of 'E-mail Battleship'.
 * Miss. Miss. You sunk my inbox!
 *
 * A legitimate license entitles you to support from the developers,
 * and the warm fuzzy feeling of feeding a couple of obsessed developers
 * who want to help you get more done.
 *
 \* - Jeff Standen, Darren Sugita, Dan Hildebrandt
 *	 Webgroup Media LLC - Developers of Cerb
 */
class CerberusMail {
	private function __construct() {}
	
	static function parseRfcAddresses($string, $exclude_controlled_addresses=false) {
		$results = array();
		$string = rtrim(str_replace(';',',',$string),' ,');
		@$parsed = imap_rfc822_parse_adrlist($string, 'localhost');
		
		$exclude_list = DevblocksPlatform::getPluginSetting('cerberusweb.core', CerberusSettings::PARSER_AUTO_REQ_EXCLUDE, CerberusSettingsDefaults::PARSER_AUTO_REQ_EXCLUDE);
		@$excludes = DevblocksPlatform::parseCrlfString($exclude_list);
		
		if(is_array($parsed))
		foreach($parsed as $parsed_addy) {
			@$mailbox = strtolower($parsed_addy->mailbox);
			@$host = strtolower($parsed_addy->host);
			@$personal = isset($parsed_addy->personal) ? $parsed_addy->personal : null;
			
			if(empty($mailbox) || empty($host))
				continue;
			
			if(0 == strcasecmp($mailbox, 'invalid_address'))
				continue;
			
			if(0 == strcasecmp($host, '.syntax-error.'))
				continue;
			
			// Are we excluding Cerb controlled addresses?
			if($exclude_controlled_addresses) {
				$check_address = $mailbox.'@'.$host;
				
				// If this is a local address and we're excluding them, skip it
				if(DAO_AddressOutgoing::isLocalAddress($check_address))
					continue;
				
				$skip = false;
				
				// Filter explicit excludes
				if(is_array($excludes) && !empty($excludes))
				foreach($excludes as $excl_pattern) {
					if(@preg_match(DevblocksPlatform::parseStringAsRegExp($excl_pattern), $check_address))
						$skip = true;
				}
				
				if($skip)
					continue;
			}
			
			$results[$mailbox . '@' . $host] = array(
				'full_email' => !empty($personal) ? imap_rfc822_write_address($mailbox, $host, $personal) : imap_rfc822_write_address($mailbox, $host, null),
				'email' => $mailbox . '@' . $host,
				'mailbox' => $mailbox,
				'host' => $host,
				'personal' => $personal,
			);
		}
		
		@imap_errors();
		
		return $results;
	}
	
	static private function _parseCustomHeaders(array $headers) {
		if(!is_array($headers))
			return array();
		
		$results = array();
		
		foreach($headers as $header) {
			@list($name, $value) = explode(':', $header);
			
			$name = trim(strtolower($name));
			$value = trim($value);
			
			$results[$name] = $value;
		}
		
		return $results;
	}
	
	static function quickSend($to, $subject, $body, $from_addy=null, $from_personal=null, $custom_headers=array(), $format=null, $html_template_id=null, $file_ids=array()) {
		try {
			$mail_service = DevblocksPlatform::getMailService();
			$mail = $mail_service->createMessage();
	
			$settings = DevblocksPlatform::getPluginSettingsService();
			
			if(empty($from_addy) || empty($from_personal)) {
				if(null == ($replyto_default = DAO_AddressOutgoing::getDefault()))
					throw new Exception("There is no default reply-to.");
				
				if(empty($from_addy))
					$from_addy = $replyto_default->email;
				if(empty($from_personal))
					$from_personal = $replyto_default->getReplyPersonal();
			}
			
			$mail->setTo(DevblocksPlatform::parseCsvString($to));

			$custom_headers = self::_parseCustomHeaders($custom_headers);

			// If we have a custom from, override the sender info
			if(isset($custom_headers['from'])) {
				if(false != ($custom_froms = mailparse_rfc822_parse_addresses($custom_headers['from']))) {
					$from_addy = $custom_froms[0]['address'];
					$from_personal = ($custom_froms[0]['display'] != $custom_froms[0]['address']) ? $custom_froms[0]['display'] : null;
				}
				
				unset($custom_headers['from']);
			}
			
			if(!empty($from_personal)) {
				$mail->setFrom($from_addy, $from_personal);
			} else {
				$mail->setFrom($from_addy);
			}
			
			// If we have a custom subject, use it instead
			if(isset($custom_headers['subject'])) {
				$mail->setSubject($custom_headers['subject']);
				unset($custom_headers['subject']);
				
			} else {
				$mail->setSubject($subject);
			}
			
			// Generate a message-id
			$mail->generateId();
			
			$headers = $mail->getHeaders();
			
			$headers->addTextHeader('X-Mailer','Cerb ' . APP_VERSION . ' (Build '.APP_BUILD.')');
			
			// Add custom headers
			
			if(is_array($custom_headers) && !empty($custom_headers))
			foreach($custom_headers as $header_key => $header_val) {
				if(!empty($header_key) && !empty($header_val)) {
					if($headers->has($header_key))
						$headers->removeAll($header_key);
					
					$headers->addTextHeader(mb_convert_case($header_key, MB_CASE_TITLE), $header_val);
				}
			}
			
			// Body
			
			switch($format) {
				case 'parsedown':
					self::_generateBodyMarkdown($mail, $body, null, null, $html_template_id);
					break;
					
				default:
					$mail->setBody($body);
					break;
			}
			
			// Attachments
			
			if(!empty($file_ids) && is_array($file_ids)) {
				foreach($file_ids as $file_id) {
					// Attach the file
					if(false != ($attachment = DAO_Attachment::get($file_id))) {
						if(false !== ($fp = DevblocksPlatform::getTempFile())) {
							if(false !== $attachment->getFileContents($fp)) {
								$attach = Swift_Attachment::fromPath(DevblocksPlatform::getTempFileInfo($fp), $attachment->mime_type);
								$attach->setFilename($attachment->display_name);
								$mail->attach($attach);
								fclose($fp);
							}
						}
					}
				}
			}
			
			// [TODO] Report when the message wasn't sent.
			// [TODO] We can use '$failedRecipients' for this
			if(!$mail_service->send($mail)) {
				return false;
			}
			
		} catch (Exception $e) {
			return false;
		}
		
		return true;
	}

	static function compose($properties) {
		/*
		 'group_id'
		 'bucket_id'
		 'worker_id'
		 'org_id'
		 'to'
		 'cc'
		 'bcc'
		 'subject'
		 'content'
		 'content_format'
		 'html_template_id'
		 'files'
		 'forward_files'
		 'closed'
		 'ticket_reopen'
		 'dont_send'
		 'draft_id'
		 */
		
		@$group_id = $properties['group_id'];
		@$bucket_id = intval($properties['bucket_id']);
		
		@$is_broadcast = intval($properties['is_broadcast']);
		
		@$worker_id = $properties['worker_id'];
		$worker = null;
		
		if(empty($worker_id) && null != ($worker = CerberusApplication::getActiveWorker()))
			$worker_id = $worker->id;

		// Worker
		if(!empty($worker_id))
			$worker = DAO_Worker::get($worker_id);
		
		// Group
		if(null == ($group = DAO_Group::get($group_id)))
			return;
		
		// Bucket
		if(!$bucket_id || false == ($bucket = DAO_Bucket::get($bucket_id)) || $bucket->group_id != $group->id)
			$bucket = $group->getDefaultBucket();
		
		// Changing the outgoing message through a VA (global)
		Event_MailBeforeSent::trigger($properties, null, null, $group_id);
		
		// Changing the outgoing message through a VA (group)
		Event_MailBeforeSentByGroup::trigger($properties, null, null, $group_id);
		
		@$org_id = $properties['org_id'];
		@$toStr = $properties['to'];
		@$cc = $properties['cc'];
		@$bcc = $properties['bcc'];
		@$subject = $properties['subject'];
		@$content = $properties['content'];
		@$content_format = $properties['content_format'];
		@$html_template_id = $properties['html_template_id'];
		@$files = $properties['files'];
		@$embedded_files = array();
		@$forward_files = $properties['forward_files'];
		
		@$closed = $properties['closed'];
		@$ticket_reopen = $properties['ticket_reopen'];
		
		$from_replyto = $group->getReplyTo($bucket->id);
		$personal = $group->getReplyPersonal($bucket->id, $worker);
		
		$mask = CerberusApplication::generateTicketMask();

		if(empty($subject)) $subject = '(no subject)';
		
		// add mask to subject if group setting calls for it
		@$group_has_subject = intval(DAO_GroupSettings::get($group_id,DAO_GroupSettings::SETTING_SUBJECT_HAS_MASK,0));
		@$group_subject_prefix = DAO_GroupSettings::get($group_id,DAO_GroupSettings::SETTING_SUBJECT_PREFIX,'');
		$prefix = sprintf("[%s#%s] ",
			!empty($group_subject_prefix) ? ($group_subject_prefix.' ') : '',
			$mask
		);
		$subject_mailed = (sprintf('%s%s',
			$group_has_subject ? $prefix : '',
			$subject
		));
		
		// [JAS]: Replace any semi-colons with commas (people like using either)
		$toList = CerberusMail::parseRfcAddresses($toStr);
		
		$mail_headers = array();
		$mail_headers['X-CerberusCompose'] = '1';
		
		try {
			$mail_service = DevblocksPlatform::getMailService();
			$email = $mail_service->createMessage();

			// To
			if(is_array($toList))
			foreach($toList as $k => $v) {
				if(!empty($v['personal'])) {
					$email->addTo($k, $v['personal']);
				} else {
					$email->addTo($k);
				}
			}
			
			// Cc
			$ccList = CerberusMail::parseRfcAddresses($cc);
			if(is_array($ccList) && !empty($ccList)) {
				foreach($ccList as $k => $v) {
					if(!empty($v['personal'])) {
						$email->addCc($k, $v['personal']);
					} else {
						$email->addCc($k);
					}
				}
			}
			
			// Bcc
			$bccList = CerberusMail::parseRfcAddresses($bcc);
			if(is_array($bccList) && !empty($bccList)) {
				foreach($bccList as $k => $v) {
					if(!empty($v['personal'])) {
						$email->addBcc($k, $v['personal']);
					} else {
						$email->addBcc($k);
					}
				}
			}
			
			if(!empty($personal)) {
				$email->setFrom($from_replyto->email, $personal);
			} else {
				$email->setFrom($from_replyto->email);
			}
			
			$email->setSubject($subject_mailed);
			$email->generateId();
			
			$headers = $email->getHeaders();
			
			$headers->addTextHeader('X-Mailer','Cerb ' . APP_VERSION . ' (Build '.APP_BUILD.')');
			
			// Custom headers
			
			if(isset($properties['headers']) && is_array($properties['headers']))
			foreach($properties['headers'] as $header_key => $header_val) {
				if(!empty($header_key) && is_string($header_key) && is_string($header_val)) {
					if(NULL == ($header = $headers->get($header_key))) {
						$headers->addTextHeader($header_key, $header_val);
					} else {
						$header->setValue($header_val);
					}
				}
			}
			
			// Body
			
			switch($content_format) {
				case 'parsedown':
					$embedded_files = self::_generateBodyMarkdown($email, $content, $group_id, $bucket->id, $html_template_id);
					break;
					
				default:
					$email->setBody($content);
					break;
			}
			
			// Mime Attachments
			if (is_array($files) && !empty($files)) {
				foreach ($files['tmp_name'] as $idx => $file) {
					if(empty($file) || empty($files['name'][$idx]))
						continue;
	
					$email->attach(Swift_Attachment::fromPath($file)->setFilename($files['name'][$idx]));
				}
			}
			
			// Forward Attachments
			if(!empty($forward_files) && is_array($forward_files)) {
				foreach($forward_files as $file_id) {
					if(false != ($attachment = DAO_Attachment::get($file_id))) {
						if(false !== ($fp = DevblocksPlatform::getTempFile())) {
							if(false !== $attachment->getFileContents($fp)) {
								$attach = Swift_Attachment::fromPath(DevblocksPlatform::getTempFileInfo($fp), $attachment->mime_type);
								$attach->setFilename($attachment->display_name);
								$email->attach($attach);
								fclose($fp);
							}
						}
					}
				}
			}
			
			// Headers
			foreach($email->getHeaders()->getAll() as $hdr) {
				if(null != ($hdr_val = $hdr->getFieldBody())) {
					if(!empty($hdr_val))
						$mail_headers[$hdr->getFieldName()] = CerberusParser::fixQuotePrintableString($hdr_val, LANG_CHARSET_CODE);
				}
			}
			
			if(!empty($toList) && (!isset($properties['dont_send']) || empty($properties['dont_send']))) {
				if(!$mail_service->send($email)) {
					throw new Exception('Mail failed to send: unknown reason');
				}
			}
	
		} catch (Exception $e) {
			@$draft_id = $properties['draft_id'];
			
			if(empty($draft_id)) {
				$params = array(
					'to' => $toStr,
					'group_id' => $group->id,
					'bucket_id' => $bucket->id,
				);
				
				if(!empty($cc))
					$params['cc'] = $cc;
					
				if(!empty($bcc))
					$params['bcc'] = $bcc;
					
				$fields = array(
					DAO_MailQueue::TYPE => Model_MailQueue::TYPE_COMPOSE,
					DAO_MailQueue::TICKET_ID => 0,
					DAO_MailQueue::WORKER_ID => intval($worker_id),
					DAO_MailQueue::UPDATED => time()+5, // small offset
					DAO_MailQueue::HINT_TO => $toStr,
					DAO_MailQueue::SUBJECT => $subject,
					DAO_MailQueue::BODY => $content,
					DAO_MailQueue::PARAMS_JSON => json_encode($params),
					DAO_MailQueue::IS_QUEUED => !empty($worker) ? 0 : 1,
					DAO_MailQueue::QUEUE_DELIVERY_DATE => time(),
				);
				DAO_MailQueue::create($fields);
			}
			
			return false;
		}
		
		$fromAddressInst = CerberusApplication::hashLookupAddress($from_replyto->email, true);
		$fromAddressId = $fromAddressInst->id;
		
		// [TODO] this is redundant with the Parser code.  Should be refactored later
		
		// Organization ID from first requester
		if(empty($org_id)) {
			reset($toList);
			if(null != ($first_req = DAO_Address::lookupAddress(key($toList),true))) {
				if(!empty($first_req->contact_org_id))
					$org_id = $first_req->contact_org_id;
			}
		}
		
		$fields = array(
			DAO_Ticket::MASK => $mask,
			DAO_Ticket::SUBJECT => $subject,
			DAO_Ticket::CREATED_DATE => time(),
			DAO_Ticket::FIRST_WROTE_ID => $fromAddressId,
			DAO_Ticket::LAST_WROTE_ID => $fromAddressId,
			DAO_Ticket::ORG_ID => intval($org_id),
			DAO_Ticket::LAST_ACTION_CODE => CerberusTicketActionCode::TICKET_WORKER_REPLY,
			DAO_Ticket::IMPORTANCE => 50,
		);
		
		// "Next:" [TODO] This is highly redundant with CerberusMail::reply
		
		if(isset($properties['owner_id'])) {
			$fields[DAO_Ticket::OWNER_ID] = intval($properties['owner_id']);
		}
		
		if(isset($ticket_reopen) && !empty($ticket_reopen)) {
			if(is_numeric($ticket_reopen) && false != (@$reopen_at = strtotime('now', $ticket_reopen))) {
				$fields[DAO_Ticket::REOPEN_AT] = $reopen_at;
			} else if(false !== (@$reopen_at = strtotime($ticket_reopen))) {
				$fields[DAO_Ticket::REOPEN_AT] = $reopen_at;
			}
		}

		// End "Next:"
		
		$ticket_id = DAO_Ticket::create($fields);

		$fields = array(
			DAO_Message::TICKET_ID => $ticket_id,
			DAO_Message::CREATED_DATE => time(),
			DAO_Message::ADDRESS_ID => $fromAddressId,
			DAO_Message::IS_OUTGOING => 1,
			DAO_Message::WORKER_ID => intval($worker_id),
			DAO_Message::IS_BROADCAST => $is_broadcast ? 1 : 0,
			DAO_Message::IS_NOT_SENT => @$properties['dont_send'] ? 1 : 0,
		);
		$message_id = DAO_Message::create($fields);
		
		// Content
		Storage_MessageContent::put($message_id, $content);

		// Set recipients to requesters
		foreach($toList as $to_addy => $to_data) {
			DAO_Ticket::createRequester($to_addy, $ticket_id);
		}
		
		// Headers
		DAO_MessageHeader::creates($message_id, $mail_headers);
		
		// add files to ticket
		if (is_array($files) && !empty($files)) {
			reset($files);
			foreach ($files['tmp_name'] as $idx => $file) {
				if(empty($file) || empty($files['name'][$idx]) || !file_exists($file))
					continue;

				// Dupe detection
				@$sha1_hash = sha1_file($file, false);
				
				if(false == ($file_id = DAO_Attachment::getBySha1Hash($sha1_hash, $files['name'][$idx]))) {
					$fields = array(
						DAO_Attachment::DISPLAY_NAME => $files['name'][$idx],
						DAO_Attachment::MIME_TYPE => $files['type'][$idx],
						DAO_Attachment::STORAGE_SHA1HASH => $sha1_hash,
					);
					$file_id = DAO_Attachment::create($fields);
					
					// Content
					if(null !== ($fp = fopen($file, 'rb'))) {
						Storage_Attachments::put($file_id, $fp);
						fclose($fp);
					}
				}

				// Link
				if($file_id)
					DAO_AttachmentLink::create($file_id, CerberusContexts::CONTEXT_MESSAGE, $message_id);
				
				@unlink($file);
			}
		}

		// Forwarded attachments
		if(isset($properties['link_forward_files']) && !empty($properties['link_forward_files'])) {
			// Attachments
			if(is_array($forward_files) && !empty($forward_files)) {
				DAO_AttachmentLink::addLinks(CerberusContexts::CONTEXT_MESSAGE, $message_id, $forward_files);
			}
		}
		
		// Link embedded files
		if(isset($embedded_files) && is_array($embedded_files) && !empty($embedded_files)) {
			DAO_AttachmentLink::addLinks(CerberusContexts::CONTEXT_MESSAGE, $message_id, $embedded_files);
		}
		
		// Finalize ticket
		$fields = array(
			DAO_Ticket::FIRST_MESSAGE_ID => $message_id,
			DAO_Ticket::LAST_MESSAGE_ID => $message_id,
		);
		
		if(isset($closed) && 1==$closed)
			$fields[DAO_Ticket::IS_CLOSED] = 1;
		if(isset($closed) && 2==$closed)
			$fields[DAO_Ticket::IS_WAITING] = 1;
		
		// Move last, so the event triggers properly
		$fields[DAO_Ticket::GROUP_ID] = $group->id;
		$fields[DAO_Ticket::BUCKET_ID] = $bucket->id;
		
		DAO_Ticket::update($ticket_id, $fields);
		
		// Train as not spam
		CerberusBayes::markTicketAsNotSpam($ticket_id);
		
		// Custom fields
		@$custom_fields = isset($properties['custom_fields']) ? $properties['custom_fields'] : array();
		if(is_array($custom_fields) && !empty($custom_fields)) {
			DAO_CustomFieldValue::formatAndSetFieldValues(CerberusContexts::CONTEXT_TICKET, $ticket_id, $custom_fields);
			
			// Link any fieldsets
			DAO_CustomFieldset::linkToContextByFieldIds(CerberusContexts::CONTEXT_TICKET, $ticket_id, array_keys($custom_fields));
		}
		
		// Events
		if(!empty($message_id) && !empty($group_id)) {
			// After message sent (global)
			Event_MailAfterSent::trigger($message_id, $group_id);
			
			// After message sent in group
			Event_MailAfterSentByGroup::trigger($message_id, $group_id);

			// Mail received
			Event_MailReceived::trigger($message_id);
			
			// Mail received by group
			Event_MailReceivedByGroup::trigger($message_id, $group_id);
		}
		
		return intval($ticket_id);
	}
	
	static function sendTicketMessage($properties=array()) {
		$settings = DevblocksPlatform::getPluginSettingsService();
		
		/*
		'draft_id'
		'message_id'
		'is_forward'
		'is_broadcast'
		'subject'
		'to'
		'cc'
		'bcc'
		'content'
		'content_format' // markdown, parsedown, html
		'html_template_id'
		'headers'
		'files'
		'forward_files'
		'link_forward_files'
		'closed'
		'ticket_reopen'
		'group_id'
		'bucket_id'
		'owner_id'
		'worker_id'
		'is_autoreply'
		'custom_fields'
		'dont_send'
		'dont_keep_copy'
		*/

		try {
			// objects
			$mail_service = DevblocksPlatform::getMailService();
			$mail = $mail_service->createMessage();
			
			@$reply_message_id = $properties['message_id'];
			
			if(null == ($message = DAO_Message::get($reply_message_id)))
				return;
				
			$ticket_id = $message->ticket_id;
			
			// Ticket
			if(null == ($ticket = DAO_Ticket::get($ticket_id)))
				return;
				
			// Group
			if(null == ($group = DAO_Group::get($ticket->group_id)))
				return;
			
			// Changing the outgoing message through a VA (global)
			Event_MailBeforeSent::trigger($properties, $message->id, $ticket->id, $group->id);
			
			// Changing the outgoing message through a VA (group)
			Event_MailBeforeSentByGroup::trigger($properties, $message->id, $ticket->id, $group->id);
			
			// Re-read properties
			@$content = $properties['content'];
			@$content_format = $properties['content_format'];
			@$html_template_id = intval($properties['html_template_id']);
			@$files = $properties['files'];
			@$is_forward = $properties['is_forward'];
			@$is_broadcast = $properties['is_broadcast'];
			@$forward_files = $properties['forward_files'];
			@$embedded_files = array();
			@$worker_id = $properties['worker_id'];
			@$subject = $properties['subject'];
			
			@$is_autoreply = $properties['is_autoreply'];
			
			$message_headers = DAO_MessageHeader::getAll($reply_message_id);

			$from_replyto = $group->getReplyTo($ticket->bucket_id);
			$from_personal = $group->getReplyPersonal($ticket->bucket_id, $worker_id);
			
			/*
			 * If this ticket isn't spam trained
			 * and our outgoing message isn't an autoreply
			 * and a worker sent this
			 */
			if($ticket->spam_training == CerberusTicketSpamTraining::BLANK
				&& empty($is_autoreply)
				&& !empty($worker_id)) {
				CerberusBayes::markTicketAsNotSpam($ticket_id);
			}
				
			// Headers
			if(!empty($from_personal)) {
				$mail->setFrom($from_replyto->email, $from_personal);
			} else {
				$mail->setFrom($from_replyto->email);
			}
			
			$mail->generateId();
			
			$headers = $mail->getHeaders();
			
			$headers->addTextHeader('X-Mailer','Cerb ' . APP_VERSION . ' (Build '.APP_BUILD.')');
	
			// Subject
			if(empty($subject)) $subject = $ticket->subject;
			
			if(!empty($is_forward)) { // forward
				$mail->setSubject($subject);
				
			} else { // reply
				@$group_has_subject = intval(DAO_GroupSettings::get($ticket->group_id, DAO_GroupSettings::SETTING_SUBJECT_HAS_MASK,0));
				@$group_subject_prefix = DAO_GroupSettings::get($ticket->group_id, DAO_GroupSettings::SETTING_SUBJECT_PREFIX,'');
				
				$prefix = sprintf("[%s#%s] ",
					!empty($group_subject_prefix) ? ($group_subject_prefix.' ') : '',
					$ticket->mask
				);
				
				$mail->setSubject(sprintf('Re: %s%s',
					$group_has_subject ? $prefix : '',
					$subject
				));
			}
			
			// References
			if(!empty($message) && false !== (@$in_reply_to = $message_headers['message-id'])) {
				$headers->addTextHeader('References', $in_reply_to);
				$headers->addTextHeader('In-Reply-To', $in_reply_to);
			}
	
			// Default requester reply
			if(empty($properties['to']) && !$is_forward) {
				// Auto-reply handling (RFC-3834 compliant)
				if(!empty($is_autoreply))
					$headers->addTextHeader('Auto-Submitted','auto-replied');
				
				// Recipients
				$requesters = DAO_Ticket::getRequestersByTicket($ticket_id);
				
				if(is_array($requesters))
				foreach($requesters as $requester) { /* @var $requester Model_Address */
					$first_email = strtolower($requester->email);
					$first_split = explode('@', $first_email);
			
					if(!is_array($first_split) || count($first_split) != 2)
						continue;
			
					// Ourselves?
					if(DAO_AddressOutgoing::isLocalAddressId($requester->id))
						continue;

					if($is_autoreply) {
						// If return-path is blank
						if(isset($message_headers['return-path']) && $message_headers['return-path'] == '<>')
							continue;
						
						// Ignore autoresponses to autoresponses
						if(isset($message_headers['auto-submitted']) && $message_headers['auto-submitted'] != 'no')
							continue;
	
						// Bulk mail?
						if(isset($message_headers['precedence']) &&
							($message_headers['precedence'] == 'list' || $message_headers['precedence'] == 'junk' || $message_headers['precedence'] == 'bulk'))
							continue;
					}
						
					// Ignore bounces
					if($first_split[0] == "postmaster" || $first_split[0] == "mailer-daemon")
						continue;
						
					// Auto-reply just to the initial requester
					$mail->addTo($requester->email);
				}
				
			// Forward or overload
			} elseif(!empty($properties['to'])) {
				// To
				$aTo = CerberusMail::parseRfcAddresses($properties['to']);
				if(is_array($aTo))
				foreach($aTo as $k => $v) {
					if(!empty($v['personal'])) {
						$mail->addTo($k, $v['personal']);
					} else {
						$mail->addTo($k);
					}
				}
			}
			
			// Ccs
			if(!empty($properties['cc'])) {
				$aCc = CerberusMail::parseRfcAddresses($properties['cc']);
				if(is_array($aCc))
				foreach($aCc as $k => $v) {
					if(!empty($v['personal'])) {
						$mail->addCc($k, $v['personal']);
					} else {
						$mail->addCc($k);
					}
				}
			}
			
			// Bccs
			if(!empty($properties['bcc'])) {
				$aBcc = CerberusMail::parseRfcAddresses($properties['bcc']);
				if(is_array($aBcc))
				foreach($aBcc as $k => $v) {
					if(!empty($v['personal'])) {
						$mail->addBcc($k, $v['personal']);
					} else {
						$mail->addBcc($k);
					}
				}
			}
			
			// Custom headers
			
			if(isset($properties['headers']) && is_array($properties['headers']))
			foreach($properties['headers'] as $header_key => $header_val) {
				if(!empty($header_key) && is_string($header_key) && is_string($header_val)) {
					if(NULL == ($header = $headers->get($header_key))) {
						$headers->addTextHeader($header_key, $header_val);
					} else {
						$header->setValue($header_val);
					}
				}
			}
			
			// Body
			
			switch($content_format) {
				case 'parsedown':
					$embedded_files = self::_generateBodyMarkdown($mail, $content, $ticket->group_id, $ticket->bucket_id, $html_template_id);
					break;
					
				default:
					$mail->setBody($content);
					break;
			}
	
			// Mime Attachments
			if (is_array($files) && !empty($files)) {
				if(isset($files['tmp_name']))
				foreach($files['tmp_name'] as $idx => $file) {
					if(empty($file) || empty($files['name'][$idx]))
						continue;
	
					$mail->attach(Swift_Attachment::fromPath($file)->setFilename($files['name'][$idx]));
				}
			}
	
			// Forward Attachments
			if(!empty($forward_files) && is_array($forward_files)) {
				foreach($forward_files as $file_id) {
					// Attach the file
					if(false != ($attachment = DAO_Attachment::get($file_id))) {
						if(false !== ($fp = DevblocksPlatform::getTempFile())) {
							if(false !== $attachment->getFileContents($fp)) {
								$attach = Swift_Attachment::fromPath(DevblocksPlatform::getTempFileInfo($fp), $attachment->mime_type);
								$attach->setFilename($attachment->display_name);
								$mail->attach($attach);
								fclose($fp);
							}
						}
					}
				}
			}

			// Send
			$recipients = $mail->getTo();
			$send_headers = array();
			
			// Save headers before sending
			foreach($headers->getAll() as $hdr) {
				if(null != ($hdr_val = $hdr->getFieldBody())) {
					if(!empty($hdr_val))
						$send_headers[$hdr->getFieldName()] = CerberusParser::fixQuotePrintableString($hdr_val, LANG_CHARSET_CODE);
				}
			}
			
			// If blank recipients or we're not supposed to send
			if(empty($recipients) || (isset($properties['dont_send']) && $properties['dont_send'])) {
				// ...do nothing
				
			} else { // otherwise send
				if(!$mail_service->send($mail)) {
					throw new Exception('Mail not sent.');
				}
			}
			
		} catch (Exception $e) {
			@$draft_id = $properties['draft_id'];

			// Only if we weren't trying to send a draft already...
			if(empty($draft_id)) {
				$params = array(
					'in_reply_message_id' => $properties['message_id'],
				);
				
				if(isset($properties['cc']))
					$params['cc'] = $properties['cc'];
					
				if(isset($properties['bcc']))
					$params['bcc'] = $properties['bcc'];
					
				if(!empty($is_autoreply))
					$params['is_autoreply'] = true;
				
				if(empty($to)) {
					$hint_to = '(requesters)';
				} else {
					$hint_to = implode(', ', array_keys($mail->getTo()));
				}
				
				$fields = array(
					DAO_MailQueue::TYPE => empty($is_forward) ? Model_MailQueue::TYPE_TICKET_REPLY : Model_MailQueue::TYPE_TICKET_FORWARD,
					DAO_MailQueue::TICKET_ID => $properties['ticket_id'],
					DAO_MailQueue::WORKER_ID => intval($worker_id),
					DAO_MailQueue::UPDATED => time()+5, // small offset
					DAO_MailQueue::HINT_TO => $hint_to,
					DAO_MailQueue::SUBJECT => $subject,
					DAO_MailQueue::BODY => $properties['content'],
					DAO_MailQueue::PARAMS_JSON => json_encode($params),
					DAO_MailQueue::IS_QUEUED => empty($worker_id) ? 1 : 0,
					DAO_MailQueue::QUEUE_DELIVERY_DATE => time(),
				);
				$draft_id = DAO_MailQueue::create($fields);
			}
			
			$last_error_message = $mail_service->getLastErrorMessage();
			
			if($e instanceof Swift_TransportException && !$last_error_message) {
				$last_error_message = $e->getMessage();
			}
			
			// If we have an error message, log it on the draft
			if($draft_id && !empty($last_error_message)) {
				$fields = array(
					DAO_Comment::OWNER_CONTEXT => CerberusContexts::CONTEXT_APPLICATION,
					DAO_Comment::OWNER_CONTEXT_ID => 0,
					DAO_Comment::CONTEXT => CerberusContexts::CONTEXT_DRAFT,
					DAO_Comment::CONTEXT_ID => $draft_id,
					DAO_Comment::COMMENT => 'Error sending message: ' . $last_error_message,
					DAO_Comment::CREATED => time(),
				);
				DAO_Comment::create($fields);
			}
			
			return false;
		}
		
		$change_fields = array();
		
		$fromAddressInst = CerberusApplication::hashLookupAddress($from_replyto->email, true);
		$fromAddressId = $fromAddressInst->id;
		
		if((!isset($properties['dont_keep_copy']) || !$properties['dont_keep_copy'])
			&& empty($is_autoreply)) {
			$change_fields[DAO_Ticket::LAST_MESSAGE_ID] = $fromAddressId;
			$change_fields[DAO_Ticket::LAST_WROTE_ID] = $fromAddressId;
			$change_fields[DAO_Ticket::UPDATED_DATE] = time();
			
			if(!empty($worker_id)) {
				$change_fields[DAO_Ticket::LAST_ACTION_CODE] = CerberusTicketActionCode::TICKET_WORKER_REPLY;
			}
			
			// Only change the subject if not forwarding
			if(!empty($subject) && !$is_forward) {
				$change_fields[DAO_Ticket::SUBJECT] = $subject;
			}

			// Response time
			
			$response_epoch = ($ticket->created_date > $message->created_date) ? $ticket->created_date : $message->created_date;
			$response_time = (!empty($worker_id) ? (time() - $response_epoch) : 0);
			
			unset($response_epoch);
			
			// Fields
			
			$fields = array(
				DAO_Message::TICKET_ID => $ticket_id,
				DAO_Message::CREATED_DATE => time(),
				DAO_Message::ADDRESS_ID => $fromAddressId,
				DAO_Message::IS_OUTGOING => 1,
				DAO_Message::WORKER_ID => (!empty($worker_id) ? $worker_id : 0),
				DAO_Message::RESPONSE_TIME => $response_time,
				DAO_Message::IS_BROADCAST => $is_broadcast ? 1 : 0,
				DAO_Message::IS_NOT_SENT => @$properties['dont_send'] ? 1 : 0,
			);
			$message_id = DAO_Message::create($fields);
			
			// Store ticket.last_message_id
			$change_fields[DAO_Ticket::LAST_MESSAGE_ID] = $message_id;
			
			// First outgoing message?
			if(empty($ticket->first_outgoing_message_id) && !empty($worker_id)) {
				$change_fields[DAO_Ticket::FIRST_OUTGOING_MESSAGE_ID] = $message_id;
				$change_fields[DAO_Ticket::ELAPSED_RESPONSE_FIRST] = $response_time;
			}
			
			// Content
			Storage_MessageContent::put($message_id, $content);

			// Save cached headers
			DAO_MessageHeader::creates($message_id, $send_headers);
			
			// Attachments
			if (is_array($files) && !empty($files)) {
				reset($files);
				if(isset($files['tmp_name']))
				foreach ($files['tmp_name'] as $idx => $file) {
					if(empty($file) || empty($files['name'][$idx]) || !file_exists($file))
						continue;

					// Dupe detection
					@$sha1_hash = sha1_file($file, false);
					
					if(false == ($file_id = DAO_Attachment::getBySha1Hash($sha1_hash, $files['name'][$idx]))) {
						// Create record
						$fields = array(
							DAO_Attachment::DISPLAY_NAME => $files['name'][$idx],
							DAO_Attachment::MIME_TYPE => $files['type'][$idx],
							DAO_Attachment::STORAGE_SHA1HASH => $sha1_hash,
						);
						$file_id = DAO_Attachment::create($fields);
						
						// Content
						if(null !== ($fp = fopen($file, 'rb'))) {
							Storage_Attachments::put($file_id, $fp);
							fclose($fp);
						}
					}
					
					@unlink($file);

					// Link
					if($file_id)
						DAO_AttachmentLink::create($file_id, CerberusContexts::CONTEXT_MESSAGE, $message_id);
				}
			}
			
			// Forwarded attachments
			if(isset($properties['link_forward_files']) && !empty($properties['link_forward_files'])) {
				// Attachments
				if(is_array($forward_files) && !empty($forward_files)) {
					DAO_AttachmentLink::addLinks(CerberusContexts::CONTEXT_MESSAGE, $message_id, $forward_files);
				}
			}
			
			// Link embedded files
			if(isset($embedded_files) && is_array($embedded_files) && !empty($embedded_files)) {
				DAO_AttachmentLink::addLinks(CerberusContexts::CONTEXT_MESSAGE, $message_id, $embedded_files);
			}
		}
		
		if(isset($properties['owner_id'])) {
			if(empty($properties['owner_id']) || null != (DAO_Worker::get($properties['owner_id']))) {
				$ticket->owner_id = intval($properties['owner_id']);
				$change_fields[DAO_Ticket::OWNER_ID] = $ticket->owner_id;
			}
		}
		
		// Post-Reply Change Properties

		if(isset($properties['closed'])) {
			$reopen_at = 0;
			
			// Handle reopen date
			if(isset($properties['ticket_reopen']) && !empty($properties['ticket_reopen'])) {
				if(is_numeric($properties['ticket_reopen']) && false != (@$reopen_at = strtotime('now', $properties['ticket_reopen']))) {
				} else if(false !== (@$reopen_at = strtotime($properties['ticket_reopen']))) {
				}
			}
			
			switch($properties['closed']) {
				case 0: // open
					$change_fields[DAO_Ticket::IS_WAITING] = 0;
					$change_fields[DAO_Ticket::IS_CLOSED] = 0;
					$change_fields[DAO_Ticket::IS_DELETED] = 0;
					$change_fields[DAO_Ticket::REOPEN_AT] = 0;
					break;
				case 1: // closed
					$change_fields[DAO_Ticket::IS_WAITING] = 0;
					$change_fields[DAO_Ticket::IS_CLOSED] = 1;
					$change_fields[DAO_Ticket::IS_DELETED] = 0;
					
					if(isset($properties['ticket_reopen']))
						$change_fields[DAO_Ticket::REOPEN_AT] = $reopen_at;
					break;
				case 2: // waiting
					$change_fields[DAO_Ticket::IS_WAITING] = 1;
					$change_fields[DAO_Ticket::IS_CLOSED] = 0;
					$change_fields[DAO_Ticket::IS_DELETED] = 0;
					
					if(isset($properties['ticket_reopen']))
						$change_fields[DAO_Ticket::REOPEN_AT] = $reopen_at;
					break;
			}
		}

		// Move
		if(isset($properties['group_id']) || isset($properties['bucket_id'])) {
			@$move_to_group_id = intval($properties['group_id']);
			@$move_to_bucket_id = intval($properties['bucket_id']);
			
			// Move to the new group if it exists
			if($move_to_group_id && false != ($move_to_group = DAO_Group::get($move_to_group_id)))
				$change_fields[DAO_Ticket::GROUP_ID] = $move_to_group_id;
			
			// Validate the given bucket id
			if($move_to_bucket_id 
				&& (false == ($move_to_bucket = DAO_Bucket::get($move_to_bucket_id)) 
					|| $move_to_bucket->group_id != $move_to_group->id)) {
					$move_to_bucket_id = 0;
			}
			
			// Move to the new bucket if it is an inbox, or it belongs to the group
			if(empty($move_to_bucket_id)) {
				$move_to_bucket = $move_to_group->getDefaultBucket();
				$move_to_bucket_id = $move_to_bucket->id;
			}
			
			if($move_to_bucket) {
				$change_fields[DAO_Ticket::BUCKET_ID] = $move_to_bucket->id;
			}
		}
			
		if(!empty($ticket_id) && !empty($change_fields)) {
			DAO_Ticket::update($ticket_id, $change_fields);
		}

		// Custom fields
		@$custom_fields = isset($properties['custom_fields']) ? $properties['custom_fields'] : array();
		if(is_array($custom_fields) && !empty($custom_fields)) {
			DAO_CustomFieldValue::formatAndSetFieldValues(CerberusContexts::CONTEXT_TICKET, $ticket_id, $custom_fields, true, true, false);
		}
		
		// Events
		if(!empty($message_id) && empty($no_events)) {
			// After message sent (global)
			Event_MailAfterSent::trigger($message_id, $group->id);
			
			// After message sent in group
			Event_MailAfterSentByGroup::trigger($message_id, $group->id);
			
			// Mail received
			Event_MailReceived::trigger($message_id);
			
			// New message for group
			Event_MailReceivedByGroup::trigger($message_id, $group->id);

			// Watchers
			$context_watchers = CerberusContexts::getWatchers(CerberusContexts::CONTEXT_TICKET, $ticket_id);
			
			// Include the owner
			if(!empty($ticket->owner_id) && !isset($context_watchers[$ticket->owner_id]))
				$context_watchers[$ticket->owner_id] = true;

			if(is_array($context_watchers))
			foreach(array_unique(array_keys($context_watchers)) as $watcher_id) {
				Event_MailReceivedByWatcher::trigger($message_id, $watcher_id);
			}
		}
		
		/*
		 * Log activity (ticket.message.outbound)
		 */
		$entry = array(
			//{{actor}} responded to ticket {{target}}
			'message' => 'activities.ticket.message.outbound',
			'variables' => array(
				'target' => sprintf("[%s] %s", $ticket->mask, $ticket->subject),
				),
			'urls' => array(
				'target' => sprintf("ctx://%s:%s", CerberusContexts::CONTEXT_TICKET, $ticket->mask),
				)
		);
		CerberusContexts::logActivity('ticket.message.outbound', CerberusContexts::CONTEXT_TICKET, $ticket_id, $entry);
		
		if(isset($message_id))
			return $message_id;
	}
	
	static function relay($message_id, $emails, $include_attachments = false, $content = null, $actor_context = null, $actor_context_id = null) {
		$mail_service = DevblocksPlatform::getMailService();
		$settings = DevblocksPlatform::getPluginSettingsService();

		$workers = DAO_Worker::getAll();
		$relay_spoof_from = $settings->get('cerberusweb.core', CerberusSettings::RELAY_SPOOF_FROM, CerberusSettingsDefaults::RELAY_SPOOF_FROM);
		
		if(false == ($message = DAO_Message::get($message_id)))
			return;
		
		if(false == ($ticket = DAO_Ticket::get($message->ticket_id)))
			return;

		if(false == ($group = DAO_Group::get($ticket->group_id)))
			return;
		
		if(false == ($sender = $message->getSender()))
			return;

		$sender_name = $sender->getName();
		
		$url_writer = DevblocksPlatform::getUrlService();
		$ticket_url = $url_writer->write(sprintf('c=profiles&w=ticket&mask=%s', $ticket->mask), true);

		if($relay_spoof_from) {
			$replyto = $group->getReplyTo($ticket->bucket_id);
		} else {
			// Use the default so our 'From:' is always consistent
			$replyto = DAO_AddressOutgoing::getDefault();
		}
		
		$attachment_data = ($include_attachments)
			? DAO_AttachmentLink::getLinksAndAttachments(CerberusContexts::CONTEXT_MESSAGE, $message->id)
			: array()
			;
		
		if(empty($content)) {
			$content = sprintf("## Relayed from %s\r\n".
				"## Your reply to this message will be sent to the requesters.\r\n".
				"## Instructions: http://wiki.cerbweb.com/Email_Relay\r\n".
				"##\r\n".
				"## %s%s wrote:\r\n".
				"%s",
				$ticket_url,
				(!empty($sender_name) ? ($sender_name . ' ') : ''),
				$sender->email,
				$message->getContent()
			);
		}
		
		if(is_array($emails))
		foreach($emails as $to) {
			try {
				if(false == ($to_model = DAO_AddressToWorker::getByEmail($to)))
					continue;
				
				if(false == ($worker = $to_model->getWorker()))
					continue;
				
				$mail = $mail_service->createMessage();
				
				$mail->setTo(array($to));
	
				$headers = $mail->getHeaders(); /* @var $headers Swift_Mime_Header */
	
				if($relay_spoof_from) {
					$mail->setFrom($sender->email, !empty($sender_name) ? $sender_name : null);
					$mail->setReplyTo($replyto->email);
					
				} else {
					$replyto_personal = $replyto->getReplyPersonal($worker);
					
					if(!empty($replyto_personal)) {
						$mail->setFrom($replyto->email, !empty($replyto_personal) ? $replyto_personal : null);
						$mail->setReplyTo($replyto->email, !empty($replyto_personal) ? $replyto_personal : null);
						
					} else {
						$mail->setFrom($replyto->email);
						$mail->setReplyTo($replyto->email);
					}
				}

				// Subject
				$subject = sprintf("[relay #%s] %s", $ticket->mask, $ticket->subject);
				$mail->setSubject($subject);
	
				$headers->removeAll('message-id');

				// Sign the message so we can verify that a future relay reply is genuine
				$sign = sha1($message->id . $worker->id . APP_DB_PASS);
				$headers->addTextHeader('Message-Id', sprintf("<%s%s%s@cerb>", mt_rand(1000,9999), $sign, dechex($message->id)));
				
				$headers->addTextHeader('X-CerberusRedirect','1');
	
				// [TODO] HTML body?
				
				$mail->setBody($content);
				
				// Files
				if(!empty($attachment_data) && isset($attachment_data['attachments']) && !empty($attachment_data['attachments'])) {
					foreach($attachment_data['attachments'] as $file_id => $file) { /* @var $file Model_Attachment */
						//if('original_message.html' == $file->display_name)
						//	continue;
						
						if(false !== ($fp = DevblocksPlatform::getTempFile())) {
							if(false !== $file->getFileContents($fp)) {
								$attach = Swift_Attachment::fromPath(DevblocksPlatform::getTempFileInfo($fp), $file->mime_type);
								$attach->setFilename($file->display_name);
								$mail->attach($attach);
								fclose($fp);
							}
						}
					}
				}
				
				$result = $mail_service->send($mail);
				unset($mail);
				
				/*
				 * Log activity (ticket.message.relay)
				 */
				$entry = array(
					//{{actor}} relayed ticket {{target}} to {{worker}} ({{worker_email}})
					'message' => 'activities.ticket.message.relay',
					'variables' => array(
						'target' => sprintf("[%s] %s", $ticket->mask, $ticket->subject),
						'worker' => $worker->getName(),
						'worker_email' => $to,
						),
					'urls' => array(
						'target' => sprintf("ctx://%s:%d", CerberusContexts::CONTEXT_TICKET, $ticket->id),
						'worker' => sprintf("ctx://%s:%d", CerberusContexts::CONTEXT_WORKER, $worker->id),
						)
				);
				CerberusContexts::logActivity('ticket.message.relay', CerberusContexts::CONTEXT_TICKET, $ticket->id, $entry, $actor_context, $actor_context_id);
				
				if(!$result)
					return false;
				
			} catch (Exception $e) {
				return false;
				
			}
		}
		
		return true;
	}
	
	static function reflect(CerberusParserModel $model, $to) {
		try {
			$message = $model->getMessage(); /* @var $message CerberusParserMessage */
			
			$mail_service = DevblocksPlatform::getMailService();
			$mail = $mail_service->createMessage();
	
			$mail->setTo(array($to));

			$headers = $mail->getHeaders();

			if(isset($message->headers['subject'])) {
				if(is_array($message->headers['subject']))
					$subject = array_shift($message->headers['subject']);
				else
					$subject = $message->headers['subject'];
				$mail->setSubject($subject);
			}
			if(isset($message->headers['message-id']))
				$headers->addTextHeader('Message-Id', $message->headers['message-id']);
			if(isset($message->headers['in-reply-to']))
				$headers->addTextHeader('In-Reply-To', $message->headers['in-reply-to']);
			if(isset($message->headers['references']))
				$headers->addTextHeader('References', $message->headers['references']);
			if(isset($message->headers['from'])) {
				$sender_addy = $model->getSenderAddressModel(); /* @var $sender_addy Model_Address */
				$sender_name = $sender_addy->getName();
				$mail->setFrom($sender_addy->email, !empty($sender_name) ? $sender_name : null);
			}
			if(isset($message->headers['return-path'])) {
				$return_path = is_array($message->headers['return-path'])
					? array_shift($message->headers['return-path'])
					: $message->headers['return-path'];
				$return_path = trim($return_path,'<>');
				$mail->setReturnPath($return_path);
			}
			if(isset($message->headers['reply-to']))
				$mail->setReplyTo($message->headers['reply-to']);
				
			$headers->addTextHeader('X-CerberusRedirect','1');

			$mail->setBody($message->body);
			
			// Files
			if(is_array($message->files))
			foreach($message->files as $file_name => $file) { /* @var $file ParserFile */
				$mail->attach(Swift_Attachment::fromPath($file->tmpname)->setFilename($file_name));
			}
		
			$result = $mail_service->send($mail);
			
			if(!$result) {
				return false;
			}
			
		} catch (Exception $e) {
			return false;
		}
	}
	
	static private function _generateBodyMarkdown(&$mail, &$content, $group_id=0, $bucket_id=0, $html_template_id=0) {
		$embedded_files = array();
		$exclude_files = array();
		
		$url_writer = DevblocksPlatform::getUrlService();
		$base_url = $url_writer->write('c=files', true) . '/';
		
		// Generate an HTML part using Parsedown
		if(false !== ($html_body = DevblocksPlatform::parseMarkdown($content))) {
			
			// Determine if we have an HTML template
			if(!$html_template_id || false == ($html_template = DAO_MailHtmlTemplate::get($html_template_id))) {
				if(false == ($group = DAO_Group::get($group_id)) || false == ($html_template = $group->getReplyHtmlTemplate($bucket_id)))
					$html_template = null;
			}
			
			// Use an HTML template wrapper if we have one
			if($html_template instanceof Model_MailHtmlTemplate) {
				$tpl_builder = DevblocksPlatform::getTemplateBuilder();
				
				$html_body = $tpl_builder->build(
					$html_template->content,
					array(
						'message_body' => $html_body
					)
				);
				
				// Load the attachment links from the HTML template
				$exclude_files = array_keys($html_template->getAttachments());
			}
			
			// Purify the HTML and inline the CSS
			$html_body = DevblocksPlatform::purifyHTML($html_body, true);
			
			// Replace links with cid: in HTML part
			try {
				$html_body = preg_replace_callback(
					sprintf('|(\"%s(.*?)\")|', preg_quote($base_url)),
					function($matches) use ($base_url, $mail, &$embedded_files, $exclude_files) {
						if(3 == count($matches)) {
							$file_parts = explode('/', $matches[2]);
							@list($file_hash, $file_name) = explode('/', $matches[2], 2);
							if($file_hash && $file_name) {
								if($file_id = DAO_Attachment::getBySha1Hash($file_hash, urldecode($file_name))) {
									if($file = DAO_Attachment::get($file_id)) {
										
										if(!in_array($file_id, $exclude_files))
											$embedded_files[] = $file_id;
										
										$cid = $mail->embed(Swift_Image::newInstance($file->getFileContents(), $file->display_name, $file->mime_type));
										return sprintf('"%s"', $cid);
									}
								}
							}
						}
						
						return $matches[0];
					},
					$html_body
				);
				
			} catch(Exception $e) {
				error_log($e->getMessage());
			}

			$mail->addPart($html_body, 'text/html');
		}
		
		// Strip some Markdown in the plaintext version
		try {
			$content = preg_replace_callback(
				sprintf('|(\!\[inline-image\]\(%s(.*?)\))|', preg_quote($base_url)),
				function($matches) use ($base_url) {
					if(3 == count($matches)) {
						@list($file_hash, $file_name) = explode('/', $matches[2], 2);
						
						if($file_hash && $file_name)
							return sprintf("[Image %s]", urldecode($file_name));
					}
					
					return $matches[0];
				},
				$content
			);
			
		} catch (Exception $e) {
			error_log($e->getMessage());
		}
		
		try {
			$content = preg_replace_callback(
				sprintf('|(\!\[Image\]\((.*?)\))|'),
				function($matches) {
					if(3 == count($matches)) {
						return sprintf("%s", $matches[2]);
					}
					
					return $matches[0];
				},
				$content
			);
			
		} catch (Exception $e) {
			error_log($e->getMessage());
		}
		
		try {
			$content = DevblocksPlatform::parseMarkdown($content);
			$content = DevblocksPlatform::stripHTML($content);
			
		} catch (Exception $e) {
			error_log($e->getMessage());
		}
			
		$mail->addPart($content, 'text/plain');
		
		return $embedded_files;
	}
};
